@@ -41,12 +41,15 @@
 
 #include <drm/drmP.h>
 
-#include <linux/version.h>
-#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)
-#include <drm/drm_crtc_helper.h>
-#else
+/*
+ * Commit fcd70cd36b9b ("drm: Split out drm_probe_helper.h")
+ * moves a number of helper function definitions from
+ * drm/drm_crtc_helper.h to a new drm_probe_helper.h.
+ */
+#if defined(NV_DRM_DRM_PROBE_HELPER_H_PRESENT)
 #include <drm/drm_probe_helper.h>
 #endif
+#include <drm/drm_crtc_helper.h>
 
 #if defined(NV_DRM_DRM_GEM_H_PRESENT)
 #include <drm/drm_gem.h>
@@ -67,10 +70,28 @@ static struct nv_drm_device *dev_list = NULL;
 static void nv_drm_output_poll_changed(struct drm_device *dev)
 {
     struct drm_connector *connector = NULL;
+    struct drm_mode_config *config = &dev->mode_config;
+#if defined(NV_DRM_CONNECTOR_LIST_ITER_PRESENT)
+    struct drm_connector_list_iter conn_iter;
+    nv_drm_connector_list_iter_begin(dev, &conn_iter);
+#endif
+    /*
+     * Here drm_mode_config::mutex has been acquired unconditionally:
+     *
+     * - In the non-NV_DRM_CONNECTOR_LIST_ITER_PRESENT case, the mutex must
+     *   be held for the duration of walking over the connectors.
+     *
+     * - In the NV_DRM_CONNECTOR_LIST_ITER_PRESENT case, the mutex must be
+     *   held for the duration of a fill_modes() call chain:
+     *     connector->funcs->fill_modes()
+     *      |-> drm_helper_probe_single_connector_modes()
+     *
+     * It is easiest to always acquire the mutext for the entire connector
+     * loop.
+     */
+    mutex_lock(&config->mutex);
 
-    mutex_lock(&dev->mode_config.mutex);
-
-    nv_drm_for_each_connector(connector, dev) {
+    nv_drm_for_each_connector(connector, &conn_iter, dev) {
 
         struct nv_drm_connector *nv_connector = to_nv_connector(connector);
 
@@ -84,7 +105,10 @@ static void nv_drm_output_poll_changed(struct drm_device *dev)
             dev->mode_config.max_width, dev->mode_config.max_height);
     }
 
-    mutex_unlock(&dev->mode_config.mutex);
+    mutex_unlock(&config->mutex);
+#if defined(NV_DRM_CONNECTOR_LIST_ITER_PRESENT)
+    nv_drm_connector_list_iter_end(&conn_iter);
+#endif
 }
 
 static struct drm_framebuffer *nv_drm_framebuffer_create(
@@ -171,11 +195,6 @@ static void
 nv_drm_init_mode_config(struct nv_drm_device *nv_dev,
                         const struct NvKmsKapiDeviceResourcesInfo *pResInfo)
 {
-#if defined(NV_DRM_ATOMIC_MODESET_NONBLOCKING_COMMIT_AVAILABLE)
-    static struct drm_mode_config_helper_funcs nv_mode_config_helper = {
-        .atomic_commit_tail = nv_drm_atomic_helper_commit_tail,
-    };
-#endif
     struct drm_device *dev = nv_dev->dev;
 
     drm_mode_config_init(dev);
@@ -206,10 +225,6 @@ nv_drm_init_mode_config(struct nv_drm_device *nv_dev,
 
     dev->mode_config.async_page_flip = false;
 
-#if defined(NV_DRM_ATOMIC_MODESET_NONBLOCKING_COMMIT_AVAILABLE)
-    dev->mode_config.helper_private = &nv_mode_config_helper;
-#endif
-
     /* Initialize output polling support */
 
     drm_kms_helper_poll_init(dev);
@@ -360,10 +375,7 @@ static int nv_drm_load(struct drm_device *dev, unsigned long flags)
 
     atomic_set(&nv_dev->enable_event_handling, true);
 
-#if !defined(NV_DRM_ATOMIC_MODESET_NONBLOCKING_COMMIT_AVAILABLE)
-    init_waitqueue_head(&nv_dev->pending_commit_queue);
-    init_waitqueue_head(&nv_dev->pending_flip_queue);
-#endif
+    init_waitqueue_head(&nv_dev->flip_event_wq);
 
     mutex_unlock(&nv_dev->lock);
 
